@@ -6,8 +6,25 @@
 
 """
 
-__version__ = "0.3.13"
+__version__ = "0.3.14"
 __license__ = 'MIT'
+'''
+__uri__ = "https://github.com/byteface/domonic"
+# https://domonic.readthedocs.io/
+__title__ = 'domonic'
+__description__ = 'Generate HTML with python 3'
+
+__all__ = (
+    '__version__',
+    '__license__',
+    '__uri__',
+    '__title__',
+    '__description__'
+)
+'''
+# TITLE = __title__
+VERSION = __version__
+# LICENSE = __license__
 
 import requests
 import re
