@@ -9,6 +9,7 @@ import unittest
 # import requests
 # from mock import patch
 
+from domonic.javascript import Object
 from domonic.javascript import Math
 from domonic.javascript import Global
 from domonic.javascript import Window
@@ -24,7 +25,57 @@ class domonicTestCase(unittest.TestCase):
 
     # domonic.javascript.Math
 
+    def test_domonic_Object(self):
+
+        o = Object()
+        print(o)
+
+        # obj = {'a': 1}
+        # copy = Object.assign({}, obj)
+        # print(copy)  # { a: 1 }
+
+        # print(Object().fromEntries())
+        arr = [['0', 'a'], ['1', 'b'], ['2', 'c']]
+        obj = Object.fromEntries(arr)
+        print(obj)
+        assert obj == {'0': "a", '1': "b", '2': "c"}
+
+        obj = {'foo': 'bar', 'baz': 42}
+        print(Object.entries(obj))
+        assert Object.entries(obj) == [['foo', 'bar'], ['baz', 42]]
+
+        # array like object
+        obj = {'0': 'a', '1': 'b', '2': 'c'}
+        assert Object.entries(obj) == [['0', 'a'], ['1', 'b'], ['2', 'c']]
+
+        # array like object with random key ordering
+        # anObj = {'100': 'a', '2': 'b', '7': 'c'}
+        # print(anObj)
+        # print(Object.entries(anObj))
+        # assert Object.entries(anObj) == [['2', 'b'], ['7', 'c'], ['100', 'a']]
+
+        # returns an empty array for any primitive type
+        assert Object.entries(100) == []
+
+        # iterate through key-value gracefully
+        # obj = {'a': 5, 'b': 7, 'c': 9}
+        # for key, value in Object.entries(obj):
+            # print(f'{key} {value}')  # "a 5", "b 7", "c 9"
+
     def test_domonic_abs(self):
+        # python -m unittest tests.test_javascript.domonicTestCase.test_domonic_abs
+
+        self.assertEqual(Math.abs('-1'), 1)
+        self.assertEqual(Math.abs(-2), 2)
+        self.assertEqual(Math.abs(None), 0)
+        self.assertEqual(Math.abs(''), 0)
+        self.assertEqual(Math.abs([]), 0)
+        self.assertEqual(Math.abs([2]), 2)
+        self.assertEqual(Math.abs([1, 2]), None)
+        self.assertEqual(Math.abs({}), None)
+        self.assertEqual(Math.abs('string'), None)
+        self.assertEqual(Math.abs(), None)
+
         self.assertEqual(100, Math.abs(-100.0))
 
     def test_domonic_LN2(self):
@@ -606,6 +657,23 @@ class domonicTestCase(unittest.TestCase):
         # searchParams.set("bin2", base64)  # "q=foo&bin=E+AXQB+A&bin2=E%2BAXQB%2BA" encodes + as %2B
         # searchParams.get("bin2")  # "E+AXQB+A"
 
+    def test_javascript_FormData(self):
+        print("test_javascript_FormData")
+        # f = form(input(_type="text", _name="test", _id="test"))
+        # d = FormData(f)
+        # print(d)
+        pass
+
+    def test_javascript_Worker(self):
+        print("test_javascript_Worker")
+        # myWorker = Worker('/worker.py')
+        # first = document.querySelector('input#number1')
+        # second = document.querySelector('input#number2')
+        # first.onchange = lambda evt : \
+        #     myWorker.postMessage([first.value, second.value])
+        #     print('Message posted to worker')
+        pass
+
 
 _intID = None
 _results = []
